private import mymath;
private import myio;
private import texture;
private import log;
private import rc;
private import camera;
private import fpstimer;
private import letter;
private import vector;
private import SDL;
private import SDL_pad;
private import opengl;
private import openglu;
private import std.string;
private import std.math;
private import readygo;
private import sphere;

Texture tex;

abstract class Pbg {
    static Vec3 mpos;
    float r, g, b;
    Vec3 pos;
    Vec3 vel;
    float life;

    // from NeHe19
    static float colors[12][3]=           // Rainbow Of Colors
    [
        [1.0f,0.5f,0.5f],[0.75f,1.0f,0.5f],[0.5f,0.5f,1.0f],[1.0f,0.75f,0.5f],
        [0.5f,1.0f,0.5f],[0.5f,1.0f,0.75f],[0.5f,1.0f,1.0f],[0.5f,0.75f,1.0f],
        [1.0f,1.0f,0.5f],[0.75f,0.5f,1.0f],[1.0f,0.5f,1.0f],[1.0f,0.5f,0.75f]
        ];

    this() { life = 1; }
    this(Vec3 p0) {
        this();
        pos = p0;
    }
    this(Vec3 p0, Vec3 v0) {
        this(p0);
        vel = v0;
    }
    this(float r0, float g0, float b0) {
        r = r0;
        g = g0;
        b = b0;
    }
    this(Vec3 p0, Vec3 v0, float r0, float g0, float b0) {
        this(p0, v0);
        r = r0;
        g = g0;
        b = b0;
    }

    abstract void proc();
    abstract void draw();
    void interact(Pbg p) {}

    float x() { return pos.x; }
    float y() { return pos.y; }
    float z() { return pos.z; }
    float x(float v) { return pos.x = v; }
    float y(float v) { return pos.y = v; }
    float z(float v) { return pos.z = v; }
    float vx() { return vel.x; }
    float vy() { return vel.y; }
    float vz() { return vel.z; }
    float vx(float v) { return vel.x = v; }
    float vy(float v) { return vel.y = v; }
    float vz(float v) { return vel.z = v; }
}

class Proton : Pbg {
    int nearCnt;

    this() {
        float[3] c = colors[rnd(12)];
        r = c[0];
        g = c[1];
        b = c[2];
        for (int i = 0; i < 3; i++) {
            pos[i] = rnd(mpos[i]);
            vel[i] = rnd(mpos[i]*0.02)-mpos[i]*0.01;
        }
    }
    void proc() {
        float vm = 5;
        foreach (inout float p, int i; pos) {
            if (vel[i] > vm) vel[i] = vm;
            else if (vel[i] < -vm) vel[i] = -vm;
            p += vel[i];
            if (p < 0) {
                p = 0;
                vel[i] = -vel[i];
                p += vel[i];
            }
            else if (pos[i] >= mpos[i]) {
                p = mpos[i];
                vel[i] = -vel[i];
                p += vel[i];
            }
        }

        float rs() { return rnd(2.0)-1.0; }
        Pbg pbg = new Powder(pos,
                             Vec(-vx+rs(), -vy+rs(), -vz+rs()),
                             r, g, b);
        Main.powderMgr.add(pbg);

        nearCnt = 0;
    }
    void draw() {
        glEnable(GL_TEXTURE_2D);
        tex.bind();

        glPushMatrix();
        glTranslatef(x, y, z);

        glColor4f(r, g, b, 0.7);

        GLfloat[] texcoord = tex.texcoord;

//        float tx = 32 + ((vx > 0) ? 0 : -vx) * 10;
//        float ty = 32 + ((vy > 0) ? 0 : -vy) * 10;
        float tx = 16;
        float ty = 16;
        glBegin(GL_QUADS);
        glTexCoord3f(texcoord[0], texcoord[1], 0);
        glVertex3f(-tx, -ty, 0);
        glTexCoord3f(texcoord[0], texcoord[3], 0);
        glVertex3f(-tx, ty, 0);
        glTexCoord3f(texcoord[2], texcoord[3], 0);
        glVertex3f(tx, ty, 0);
        glTexCoord3f(texcoord[2], texcoord[1], 0);
        glVertex3f(tx, -ty, 0);
        glEnd();

        glPopMatrix();
    }
    void interact(Pbg p) {
        float r = 1;
        Vec3 dv = p.pos - pos;
        float d = 1.0 / dv.len();
        if (d < 0.1) {
            vel += r * dv * d;
            p.vel -= r * dv * d;
        }
        else {
            r *= 0.1;
            vel -= r * dv * d;
            p.vel += r * dv * d;

            nearCnt++;
            if (nearCnt > 3 && rnd(Thunder.genRatio) == 0) {
                float vr = 3;
                Pbg p = new Thunder(pos, vel*vr, r,g,b);
                Main.thunderMgr.add(p);
            }
        }
    }
}

class Powder : Pbg {
    this(Vec3 p0, Vec3 v0, float r0, float g0, float b0) {
        super(p0, v0, r0, g0, b0);
    }
    void proc() {
        float vm = 1;
        float va = 0.2;
        foreach (inout float p, int i; pos) {
            if (vel[i] > vm) vel._[i] -= va;
            else if (vel[i] < -vm) vel._[i] += va;
            p += vel[i];
            if (p < 0) {
                p = 0;
                life = 0;
            }
            else if (p >= mpos[i]) {
                p = mpos[i];
                life = 0;
            }
        }
        life -= 0.1;
    }
    void draw() {
        glEnable(GL_TEXTURE_2D);
        tex.bind();

        glPushMatrix();
        glTranslatef(x, y, z);

        glColor4f(r, g, b, life);

        GLfloat[] texcoord = tex.texcoord;

        float tx = 8;
        float ty = 8;
        glBegin(GL_QUADS);
        glTexCoord3f(texcoord[0], texcoord[1], 0);
        glVertex3f(-tx, -ty, 0);
        glTexCoord3f(texcoord[0], texcoord[3], 0);
        glVertex3f(-tx, ty, 0);
        glTexCoord3f(texcoord[2], texcoord[3], 0);
        glVertex3f(tx, ty, 0);
        glTexCoord3f(texcoord[2], texcoord[1], 0);
        glVertex3f(tx, -ty, 0);
        glEnd();

        glPopMatrix();
    }
}

class Lines : Pbg {
    Lines pair;
    static const int HISTS = 50;
    float[HISTS] hx, hy, hz;
    float[HISTS] hr, hg, hb;
    float dr, dg, db;
    int t;
    this(Vec3 p0, Vec3 v0) {
        t = 0;
        float[3] c = colors[rnd(12)];
        r = c[0];
        g = c[1];
        b = c[2];
        super(p0, v0);
    }
    void setPair(Lines p0) {
        assert(pair is null);
        pair = p0;
    }
    void proc() {
        if (pair) {
            pair.proc();

            hr[t] = r;
            hg[t] = g;
            hb[t] = b;

            if (t == 0) {
                float[3] c = colors[rnd(12)];
                dr = (c[0]-r)/HISTS;
                dg = (c[1]-g)/HISTS;
                db = (c[2]-b)/HISTS;
            }

            r += dr;
            g += dg;
            b += db;
        }

        t++;
        if (t >= HISTS) t = 0;
        hx[t] = x;
        hy[t] = y;
        hz[t] = z;

        pos += vel;
        float rs() { return rnd(10.0) + 5; }
        for (int i = 0; i < 3; i++) {
            if (pos[i] < 0) {
                pos[i] = 0;
                vel[i] = rs();
            }
            else if (pos[i] >= mpos[i]) {
                pos[i] = mpos[i];
                vel[i] = -rs();
            }
        }
    }
    void draw() {
        if (!pair) return;

        glEnable(GL_TEXTURE_2D);
        tex.bind();
        GLfloat[] texcoord = tex.texcoord;

        glBegin(GL_LINES);
        for (int i = 0; i < HISTS; i++) {
            glColor4f(hr[i], hg[i], hb[i], 0.1);
            glTexCoord3f(texcoord[0], texcoord[1], 0);
            glVertex3f(hx[i], hy[i], hz[i]);
            glTexCoord3f(texcoord[2], texcoord[3], 0);
            glVertex3f(pair.hx[i], pair.hy[i], pair.hz[i]);
//            glVertex3f(hx[i], hy[i], 300);
//            glVertex3f(pair.hx[i], pair.hy[i], 300);
        }
        glEnd();
    }
}

class Thunder : Pbg {
    static int width;
    static float dlife;
    static int genRatio;

    float[] hx, hy, hz;
    int t;
    this(Vec3 p0, Vec3 v0, float r0, float g0, float b0) {
        t = 0;
        super(p0, v0, r0, g0, b0);
    }
    void proc() {
        t++;
        life -= dlife;
        hx ~= x;
        hy ~= y;
        hz ~= z;

        pos += vel;

        float rs() { return rnd(15)-7; }
        vel.x += rs;
        vel.y += rs;
        vel.z += rs;
    }
    void draw() {
        glEnable(GL_TEXTURE_2D);
        tex.bind();

        GLfloat[] tmp = tex.texcoord;
        GLfloat[4] texcoord;
        texcoord[0] = tmp[0]*0.9+tmp[2]*0.1;
        texcoord[1] = tmp[1]*0.66+tmp[3]*0.33;
        texcoord[2] = tmp[2]*0.9+tmp[0]*0.1;
        texcoord[3] = tmp[3]*0.66+tmp[1]*0.33;

        glColor4f(r, g, b, life);
        glBegin(GL_QUADS);
        float rs() { return rnd(width) - width/2; }
        for (int i = 0; i < t-1; i++) {
            glTexCoord3f(texcoord[0], texcoord[1], 0);
            glVertex3f(hx[i], hy[i], hz[i]);
            glTexCoord3f(texcoord[2], texcoord[1], 0);
            glVertex3f(hx[i]+rs(), hy[i]+rs(), hz[i]);
            glTexCoord3f(texcoord[2], texcoord[3], 0);
            glVertex3f(hx[i+1]+rs(), hy[i+1]+rs(), hz[i+1]);
            glTexCoord3f(texcoord[0], texcoord[3], 0);
            glVertex3f(hx[i+1], hy[i+1], hz[i+1]);
        }
        glEnd();
    }
}

class PolyLine : Pbg {
    Vec3[] hp, fp, gp;
    static float theta;
    int t;
    static float width;
    static float intensity;
    static int useTexture;
    this(Vec3 p0, Vec3 v0, float r0, float g0, float b0) {
        t = 0;
        super(p0, v0, r0, g0, b0);
    }
    void proc() {
        t++;
        life -= 0.02;
        hp ~= pos;

        static Vec3 c = { x:0, y:0, z:-1 };

        if (t > 1) {
            int tp = t - 3;
            if (tp < 0) tp = 0;
            Vec3 zv = c-pos;
            zv.normalize();
            Vec3 tv = hp[t-1]-hp[tp];
            tv.normalize();
            Vec3 dv = width * (tv.cross(zv));
            fp ~= hp[t-2] + dv;
            gp ~= hp[t-2] - dv;
        }

        pos += vel;

        float co = cos(theta);
        float si = sin(theta);
        float vtmp = co*vx + si*vy;
        vy = -si*vx + co*vy;
        vx = vtmp;
    }
    void draw() {
        if (t < 3) return;

        if (useTexture) {
            glEnable(GL_TEXTURE_2D);
            tex.bind();
        }
        else {
            glDisable(GL_TEXTURE_2D);
        }

        GLfloat[] tmp = tex.texcoord;
        GLfloat[4] texcoord;
        texcoord[0] = tmp[0]*0.9+tmp[2]*0.1;
        texcoord[1] = tmp[1]*0.66+tmp[3]*0.33;
        texcoord[2] = tmp[2]*0.9+tmp[0]*0.1;
        texcoord[3] = tmp[3]*0.66+tmp[1]*0.33;

        glColor4f(r, g, b, life*intensity);
        glBegin(GL_QUADS);
        for (int i = 0; i < fp.length-1; i++) {
            glTexCoord3f(texcoord[0], texcoord[1], 0);
            glVertex3f(fp[i].x, fp[i].y, hp[i].z);
            glTexCoord3f(texcoord[2], texcoord[1], 0);
            glVertex3f(gp[i].x, gp[i].y, hp[i].z);
            glTexCoord3f(texcoord[2], texcoord[3], 0);
            glVertex3f(gp[i+1].x, gp[i+1].y, hp[i+1].z);
            glTexCoord3f(texcoord[0], texcoord[3], 0);
            glVertex3f(fp[i+1].x, fp[i+1].y, hp[i+1].z);
        }
        glEnd();
    }
}

class Simple : Pbg {
    this(Vec3 p0, Vec3 v0, float r0, float g0, float b0) {
        super(p0, v0, r0, g0, b0);
    }
    void proc() {
        foreach (inout float p, int i; pos) {
            p += vel[i];
        }
        if (z < -500) {
            life = 0;
        }
        life -= 0.01;
    }
    void draw() {
        glEnable(GL_TEXTURE_2D);
        tex.bind();

//        if (life > 0.9) return;

        glPushMatrix();
        glTranslatef(x, y, z);

        glColor4f(r, g, b, life*0.5);

        GLfloat[] texcoord = tex.texcoord;

        float tx = 16;
        float ty = 16;
        glBegin(GL_QUADS);
        glTexCoord3f(texcoord[0], texcoord[1], 0);
        glVertex3f(-tx, -ty, 0);
        glTexCoord3f(texcoord[0], texcoord[3], 0);
        glVertex3f(-tx, ty, 0);
        glTexCoord3f(texcoord[2], texcoord[3], 0);
        glVertex3f(tx, ty, 0);
        glTexCoord3f(texcoord[2], texcoord[1], 0);
        glVertex3f(tx, -ty, 0);
        glEnd();

        glPopMatrix();
    }
}

class Firework : Pbg {
    this(Vec3 p0, Vec3 v0, int c1, int c2) {
        super(p0, v0);
        cb = c1;
        ca = c2;
    }
    void proc() {
        foreach (inout float p, int i; pos) {
            p += vel[i];
        }
        if (z < -500) {
            life = 0;
        }
        if (turn < 30) {
            vel.x *= 0.91;
            vel.y *= 0.91;
            vel.z *= 0.91;
        }
        else {
            vel.y += 0.03;
            pos.x += rnd(2.0)-1.0;
            pos.z += rnd(2.0)-1.0;
        }
        life -= rnd(0.02);
        turn++;
    }
    void draw() {
        glEnable(GL_TEXTURE_2D);
        tex.bind();

//        if (life > 0.9) return;

        glPushMatrix();
        glTranslatef(x, y, z);

        float d = (life - 0.5) * 2;
        float r = colors[cb][0]*d + colors[ca][0]*(1-d);
        float g = colors[cb][1]*d + colors[ca][1]*(1-d);
        float b = colors[cb][2]*d + colors[ca][2]*(1-d);
        if (d < 0) {
            d = -d;
            r = colors[ca][0]*(1-d) + colors[0][0]*d;
            g = colors[ca][1]*(1-d) + colors[0][1]*d;
            b = colors[ca][2]*(1-d) + colors[0][2]*d;
        }
        float alpha = life+rnd(0.4);
        if (z > 200) {
            alpha -= (z-200)*0.002;
            if (alpha < -0.2) life = 0;
        }
        glColor4f(r, g, b, alpha);

        GLfloat[] texcoord = tex.texcoord;

        float tx = 15;
        float ty = 15;
        Vec3 v = vel;
        if (v.len2() < 49) {
            v.normalize();
            v *= 7;
        }
        glLineWidth(4);
        glBegin(GL_LINES);
        glTexCoord3f(texcoord[0], texcoord[1], 0);
        glVertex3f(v.x*5, v.y*5, 0);
        glTexCoord3f(texcoord[2], texcoord[3], 0);
        glVertex3f(-v.x*5, -v.y*5, 0);
/*
        glBegin(GL_QUADS);
            glTexCoord3f(texcoord[0], texcoord[1], 0);
            glVertex3f(-tx, -ty, 0);
            glTexCoord3f(texcoord[0], texcoord[3], 0);
            glVertex3f(-tx, ty, 0);
            glTexCoord3f(texcoord[2], texcoord[3], 0);
            glVertex3f(tx, ty, 0);
            glTexCoord3f(texcoord[2], texcoord[1], 0);
            glVertex3f(tx, -ty, 0);
*/
/*
        if (vel.y < 0) {
            glTexCoord3f(texcoord[0], texcoord[1], 0);
            glVertex3f(-tx+vel.x*10, -ty+vel.y*10, 0);
            glTexCoord3f(texcoord[0], texcoord[3], 0);
            glVertex3f(-tx, ty, 0);
            glTexCoord3f(texcoord[2], texcoord[3], 0);
            glVertex3f(tx, ty, 0);
            glTexCoord3f(texcoord[2], texcoord[1], 0);
            glVertex3f(tx+vel.x*10, -ty+vel.y*10, 0);
        }
        else {
            glTexCoord3f(texcoord[0], texcoord[1], 0);
            glVertex3f(-tx, -ty, 0);
            glTexCoord3f(texcoord[0], texcoord[3], 0);
            glVertex3f(-tx+vel.x*10, ty+vel.y*10, 0);
            glTexCoord3f(texcoord[2], texcoord[3], 0);
            glVertex3f(tx+vel.x*10, ty+vel.y*10, 0);
            glTexCoord3f(texcoord[2], texcoord[1], 0);
            glVertex3f(tx, -ty, 0);
        }
*/
        glEnd();

        glPopMatrix();
    }
    int cb, ca;
    int turn;
}

class ParticleMgr {
    Pbg[][] particles;
    Pbg[][] nparticles;
    int num;
    int dx, dy, dz;
    float mx, my, mz;
    float fx, fy, fz;

    int particleNum() {
        int ret = 0;
        foreach (Pbg[] ps; particles) {
            ret += ps.length;
        }
        return ret;
    }

    int getZone(Pbg p) {
        int z = cast(int)(p.x/fx)+dx*(cast(int)(p.y/fy)+dy*cast(int)(p.z/fz));
        if (z < 0) z = 0;
        else if (z >= nparticles.length) z = nparticles.length-1;
        return z;
    }

    void proc() {
        foreach (int zone, Pbg[] ps; particles) {
            foreach (int i, Pbg p1; ps) {
                for (int j = i+1; j < ps.length; j++) {
                    p1.interact(ps[j]);
                }
            }
        }

        foreach (inout Pbg[] p; nparticles) p = null;
        foreach (int zone, Pbg[] ps; particles) {
            for (int i = 0; i < ps.length; i++) {
                Pbg p = ps[i];
                p.proc();

                if (p.life > 0) {
                    int z = getZone(p);
                    nparticles[z] ~= p;
                }
            }
        }
        foreach (int i, Pbg[] p; nparticles) particles[i] = p;
    }
    void draw() {
        foreach (Pbg[] ps; particles) {
            foreach (Pbg p; ps) {
                p.draw();
            }
        }
    }

    void add(Pbg pbg) {
        int z = getZone(pbg);
        particles[z] ~= pbg;
    }

    this(int dx0, int dy0, int dz0, float mx0, float my0, float mz0) {
        dx = dx0;
        dy = dy0;
        dz = dz0;
        mx = mx0;
        my = my0;
        mz = mz0;
        fx = mx / dx;
        fy = my / dy;
        fz = mz / dz;

        nparticles.length = particles.length = dx*dy*dz;
    }
}

Main_ Main;

class Main_ {
    static this() {
        Main = new Main_();
    }

    int screenWidth_, screenHeight_;

    ParticleMgr protonMgr;
    ParticleMgr powderMgr;
    ParticleMgr linesMgr;
    ParticleMgr thunderMgr;
    ParticleMgr explodeMgr;
    ParticleMgr simpleMgr;
    ParticleMgr fireworkMgr;
    ParticleMgr[] mgrs;

    FpsTimer timer;

    Rc rc;

	void initSDL() {
		Log.info("init SDL");

		int ret = SDL_Init(SDL_INIT_VIDEO|SDL_INIT_NOPARACHUTE);
		if (ret != 0) Log.errorSDL("cannot init SDL");

		ret = SDL_InitSubSystem(SDL_INIT_JOYSTICK);
		if (ret != 0) {
			Log.warnSDL("cannot init joystick");
		}

/*
  useAudio_ = true;
  ret = SDL_InitSubSystem(SDL_INIT_AUDIO);
  if (ret != 0) {
  Log.warnSDL("cannot init audio");
  useAudio_ = false;
  }
*/

		SDL_VideoInfo* info = SDL_GetVideoInfo();

        int bpp = rc.getInt("screenBpp");
        char[] bppMsg;

		if (bpp == -1) {
			bpp = info.vfmt.BitsPerPixel;
			Log.info("using the \"best\" bpp: " ~ tostr(bpp));
		}
		else if (bpp == 0) {
			bpp = info.vfmt.BitsPerPixel;
			Log.info("using current bpp");
		}
		else {
			Log.info("using user set bpp: " ~ tostr(bpp));
		}

		SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
		SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
		SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
		SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
		SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

        if (rc.getInt("motionBlurEnable")) {
            SDL_GL_SetAttribute( SDL_GL_ACCUM_RED_SIZE, 5 );
            SDL_GL_SetAttribute( SDL_GL_ACCUM_GREEN_SIZE, 5 );
            SDL_GL_SetAttribute( SDL_GL_ACCUM_BLUE_SIZE, 5 );
        }

		screenWidth_ = rc.getInt("screenWidth");
		screenHeight_ = rc.getInt("screenHeight");
		int screenFull = rc.getInt("screenFull");

		int flags = SDL_OPENGL;

		if (screenFull) {
			flags |= SDL_FULLSCREEN;
		}

        Log.info("set video mode " ~
                 "size = " ~ tostr(screenWidth_) ~
                 "x" ~ tostr(screenHeight_) ~
                 ", bpp = " ~ tostr(bpp) ~
                 ", flags = " ~ tostr(flags));

		if (SDL_SetVideoMode(screenWidth_, screenHeight_,
							 bpp, flags) == null)
		{
			Log.errorSDL("cannot create screen");
		}

		info = SDL_GetVideoInfo();
		Log.info("set bpp: " ~ tostr(cast(int)info.vfmt.BitsPerPixel));

        SDL_WM_SetCaption(`pbg`, null);

        Log.info("pad init");
        Pad_Init();

        Pad_LoadDefault("padrc");
	}

	void initGL() {
        float bgr = rc.getFloat("clearColR");
        float bgg = rc.getFloat("clearColG");
        float bgb = rc.getFloat("clearColB");
        if (rc.getInt("reverseColor")) {
            glClearColor(1-bgr, 1-bgg, 1-bgb, 0);

            //glBlendEquationEXT(GL_FUNC_REVERSE_SUBTRACT_EXT);
            //glBlendFunc(GL_SRC_ALPHA, GL_ONE);
        }
        else {
            glClearColor(bgr, bgg, bgb, 0);
        }

		glViewport(0, 0, screenWidth_, screenHeight_);

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		float asp = cast(float)screenWidth_ / screenHeight_;
		gluPerspective(45.0, asp, 0.1, 1000);

		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
//		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
//		glBlendFunc(GL_ONE, GL_ZERO);

//        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
//        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

		glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

		// from PARSEC47
		glDisable(GL_LIGHTING);
		glDisable(GL_CULL_FACE);
		glDisable(GL_DEPTH);
		glDisable(GL_DEPTH_TEST);
        glDisable(GL_TEXTURE_2D);

        glLineWidth(rc.getInt("bgLineWidth"));
	}

    int main(char[][] args) {
        Log.init();

        rc = new Rc();
        rc.addResourceFile(`pbg.rc`);
        initSDL();
        initGL();
        initLetter();

        int enableProton = rc.getInt("enableProton");
        int enableLines = rc.getInt("enableLines");
        int enableThunder = rc.getInt("enableThunder");
        int enableTrafficLight = rc.getInt("enableTrafficLight");
        int enableFlame = rc.getInt("enableFlame");
        int enableFirework = rc.getInt("enableFirework");
        int enableSystemInfo = rc.getInt("enableSystemInfo");

        Thunder.width = rc.getInt("thunderWidth");
        Thunder.dlife = 1.0 / rc.getInt("thunderLifetime");
        Thunder.genRatio = rc.getInt("thunderGenRatio");

        PolyLine.width = rc.getFloat("trafficLightWidth");
        PolyLine.intensity = rc.getFloat("trafficLightIntensity");
        PolyLine.useTexture = rc.getInt("trafficUseTexture");
        int trafficLightNum = rc.getInt("trafficLightNum");

        int flamePercentage = rc.getInt("flamePercentage");

        tex = new Texture(`particle.bmp`);

        Pbg.mpos.x = 480;
        Pbg.mpos.y = 480;
        Pbg.mpos.z = 300;

        protonMgr = new ParticleMgr(4, 4, 4, 480, 480, 300);
        for (int i = 0; i < 100; i++) {
            Pbg pbg = new Proton();
            protonMgr.add(pbg);
        }

        Main.powderMgr = new ParticleMgr(4, 4, 4, 480, 480, 300);

        linesMgr = new ParticleMgr(1, 1, 1, 480, 480, 300);
        float rs() { return rnd(10.0) + 5; }
        for (int i = 0; i < 20; i++) {
            Lines l1 = new Lines(Vec3.zero(), Vec3.generate(&rs));
            Lines l2 = new Lines(Vec3.zero(), Vec3.generate(&rs));
            l1.setPair(l2);
            linesMgr.add(l1);
        }

        thunderMgr = new ParticleMgr(1, 1, 1, 480, 480, 300);

        explodeMgr = new ParticleMgr(1, 1, 1, 480, 480, 300);

        simpleMgr = new ParticleMgr(1, 1, 1, 480, 480, 300);

        if (enableProton) {
            mgrs ~= protonMgr;
            mgrs ~= powderMgr;
        }
        if (enableLines) mgrs ~= linesMgr;
        if (enableThunder) mgrs ~= thunderMgr;
        if (enableTrafficLight) mgrs ~= explodeMgr;
        if (enableFlame) mgrs ~= simpleMgr;

        fireworkMgr = new ParticleMgr(1, 1, 1, 480, 480, 300);
        if (enableFirework) {
            mgrs ~= fireworkMgr;
        }
        int fireworkMax = rc.getInt("fireworkMax");
        float fireworkSize = rc.getFloat("fireworkSize");
        int fireworkSphere = rc.getInt("fireworkSphere");

        ReadyGo readygo = new ReadyGo(`scene1.bmp`);
        float readygoSizeX = rc.getInt("readygoSizeX");
        float readygoSizeY = rc.getInt("readygoSizeY");

        int cameraId = rc.getInt("cameraId");
        Camera camera;
        if (cameraId == -1) {
            camera = new Camera();
        }
        else {
            camera = new Camera(cameraId);
        }
        timer = new FpsTimer(true);
		int frameSkip = rc.getInt("frameSkip");
        if (args.length >= 2) frameSkip = 0;
        if (frameSkip) {
            int fps = rc.getInt("fps");
            timer.setFps(fps);
        }
        else {
            timer.setFrameSkip(false);
        }

        int motionBlurEnable = rc.getInt("motionBlurEnable");
        int motionBlurFrames = rc.getInt("motionBlurFrames");
        float motionBlurAccum = rc.getFloat("motionBlurAccum");
        float motionBlurMult = rc.getFloat("motionBlurMult");
        if (motionBlurEnable && frameSkip) {
            Log.warn("motion blur enabled. so force disables frame skip.");
            frameSkip = false;
            timer.setFrameSkip(false);
        }

        Uint32 startTicks = SDL_GetTicks();

        int turn;
        for (turn = 0; turn < 10000000; turn++) {
            Pad_Pump(PAD_PUMP_EVERYTHING);
            if (Pad_GetButtonPushed(PAD_END_BUTTON)) {
                break;
            }
            if (Pad_GetButtonPushed(PAD_A_BUTTON)) {
                camera.end();
            }
            if (Pad_GetButtonPushed(PAD_B_BUTTON)) {
                if (camera.stop) Log.info("camera move");
                else Log.info("camera stop");
                camera.stop = !camera.stop;
            }

            // explodeMgr
            if (enableTrafficLight) {
                float a = pow(turn,1.5)/20;
                int t = (turn/10)%80;
                if (t > 60) t = t-80;
                else if (t > 40) t = -t+40;
                else if (t > 20) t = 40-t;

                static Vec3[] p,v;
                static bool inited = false;
                static float[3][] col;
                float rv() { return cast(float)(rnd(15))/10; }
                if (!inited) {
                    inited = true;

                    col.length = trafficLightNum;
                    for (int i = 0; i < trafficLightNum; i++) {
                        col[i][] = Pbg.colors[i];
                    }

                    p.length = trafficLightNum;
                    v.length = trafficLightNum;

                    for (int i = 0; i < trafficLightNum; i++) {
                        p[i].x = 240;
                        p[i].y = 240;
                        p[i].z = 200;
                        for (int j = 0; j < 2; j++) {
                            v[i][j] = rv();
                            if (rnd(2) == 0) v[i][j] = -v[i][j];
                        }
                    }
                }

                for (int i = 0; i < trafficLightNum; i++) {
                    p[i] += v[i];
                    for (int j = 0; j < 3; j++) {
                        if (p[i][j] < 0) {
                            v[i][j] = rv();
                        }
                        else if (p[i][j] > 480) {
                            v[i][j] = -rv();
                        }
                    }
                }

                PolyLine.theta = cast(float)(t)/80.0;
                for (int i = 0; i < trafficLightNum; i++) {
//                    if (rnd(3) == 0) continue;
                    Pbg pbg = new PolyLine(Vec(p[i][0],p[i][1],200),
                                           Vec(cos(a)*10,sin(a)*10,0),
                                           col[i][0],col[i][1],col[i][2]);
                    explodeMgr.add(pbg);
                }
//            }

            // simpleMgr
                if (enableFlame) {
                    for (int i = 0; i < trafficLightNum; i++) {
                        if (rnd(100) > flamePercentage) continue;
                        Pbg pbg = new Simple(Vec(p[i][0],p[i][1],200),
                                             Vec(rnd(2.0)-1,rnd(2.0)-1,-8),
//                                         1,1,1);
                                             col[i][0],col[i][1],col[i][2]);
                        simpleMgr.add(pbg);
                    }
                }
            }

            // fireworkMgr
            if (enableFirework && rnd(40) == 0 &&
                fireworkMgr.particleNum() < fireworkMax) {
                int x = rnd(300)+90;
                int y = rnd(300)+90;
                int z = 200;
                int t = 20;
                for (int j = 0; j < 2; j++) {
                int c1 = rnd(12);
                int c2 = rnd(12);
                float a1 = rnd(2*PI);
                float a2 = rnd(2*PI);
                float a3 = rnd(2*PI);
                while (c1 == c2) c2 = rnd(12);
                float s = fireworkSize - j*8;
                void addFirework(float sx, float sy, float sz) {
                    float dx, dy, dz;
                    sx *= s;
                    sy *= s;
                    sz *= s;
                    dx = sx * cos(a1) + sy * sin(a1);
                    dy = -sx * sin(a1) + sy * cos(a1);
                    sx = dx;
                    sy = dy;
                    dy = sy * cos(a2) + sz * sin(a2);
                    dz = -sy * sin(a2) + sz * cos(a2);
                    sy = dy;
                    sz = dz;
                    dz = sz * cos(a3) + sx * sin(a3);
                    dx = -sz * sin(a3) + sx * cos(a3);
                    if (fireworkSphere || dz < -0.1) {
                        Pbg pbg = new Firework(
                            Vec(x, y, z),
                            Vec(dx, dy, dz), c1, c2);
                        fireworkMgr.add(pbg);
                    }
                }
                foreach (float[3] s; (j == 0) ? sphere.sphere300 : sphere.sphere200) {
                    addFirework(s[0], s[1], s[2]);
                }
                }
            }

            foreach (ParticleMgr m; mgrs) m.proc();
            readygo.proc();

            if (!timer.shouldSkip()) {
                glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
                glMatrixMode(GL_MODELVIEW);
                glLoadIdentity();

                if (turn < 200 && cameraId == -1) {
                    gluLookAt(240, 240, -500, 240, 240, 0, 0, -1, 0);
                }
                else {
                    camera.set();
                    glTranslatef(80, 0, 0);
                }

                foreach (ParticleMgr m; mgrs) m.draw();
                readygo.draw(readygoSizeX, readygoSizeY);

                if (enableSystemInfo) drawSystemInfo();

                // drawBox();

                if (motionBlurEnable) {
                    glAccum(GL_ACCUM, motionBlurAccum);
                    int t = turn % motionBlurFrames;
                    if (t == motionBlurFrames-1) {
                        glAccum(GL_RETURN, 1.0);
                        SDL_GL_SwapBuffers();
                        glClear(GL_ACCUM_BUFFER_BIT);
                    }
                    else {
                        glAccum(GL_MULT, motionBlurMult);
                    }
                }
                else {
                    SDL_GL_SwapBuffers();
                }
            }
            timer.wait();
//            if (frameSkip) timer.wait();
//        SDL_Delay(16);
        }

        Uint32 endTicks = SDL_GetTicks();
        printf("%d FPS.\n", turn*1000/(endTicks-startTicks));

        SDL_Quit();

        return 0;
    }

    void drawSystemInfo() {
        glLoadIdentity();
        gluLookAt(240, 240, -500, 240, 240, 0, 0, -1, 0);

        int x = -35;
        int y = 45;
        glDisable(GL_TEXTURE_2D);
        drawString(`FPS`, x,y, 10,12, 1,1,1);
        drawNum(timer.getTrueFps(), x+65,y, 10,12, 1,1,1);
        y += 15;
        drawString(`FRAMES`, x,y, 10,12, 1,1,1);
        drawNum(timer.getTurn(), x+65,y, 10,12, 1,1,1);
        y += 15;
        drawString(`DROPS`, x,y, 10,12, 1,1,1);
        drawNum(timer.getDroppedTurn(), x+65,y, 10,12, 1,1,1);
/*
        y += 15;
        drawString(`SDROPS`, x,y, 10,12, 1,1,1);
        drawNum(timer.getSystemDroppedTurn(), x+65,y, 10,12, 1,1,1);
*/
        int cnt;
        foreach (ParticleMgr m; mgrs) {
            cnt += m.particleNum;
        }
        y += 15;
        drawString(`OBJS`, x,y, 10,12, 1,1,1);
        drawNum(cnt, x+65,y, 10,12, 1,1,1);
    }

    void drawBox() {
/*
  glColor4f(1,1,1,0.8);
  glDisable(GL_TEXTURE_2D);
  glBegin(GL_LINE_LOOP);
  glVertex3f(0, 0, 0);
  glVertex3f(0, 480, 0);
  glVertex3f(480, 480, 0);
  glVertex3f(480, 0, 0);
  glEnd();
  glBegin(GL_LINE_LOOP);
  glVertex3f(0, 0, 300);
  glVertex3f(0, 480, 300);
  glVertex3f(480, 480, 300);
  glVertex3f(480, 0, 300);
  glEnd();
  glBegin(GL_LINES);
  glVertex3f(0, 0, 0);
  glVertex3f(0, 0, 300);
  glVertex3f(0, 480, 0);
  glVertex3f(0, 480, 300);
  glVertex3f(480, 480, 0);
  glVertex3f(480, 480, 300);
  glVertex3f(480, 0, 0);
  glVertex3f(480, 0, 300);
  glEnd();
*/
    }

    float targetX() {
        int n = 5;
        while (protonMgr.particles[n].length == 0) n++;
        return protonMgr.particles[n][0].x;
    }
    float targetY() {
        int n = 5;
        while (protonMgr.particles[n].length == 0) n++;
        return protonMgr.particles[n][0].y;
    }
}

int main(char[][] args) {
    return Main.main(args);
}

